import readline
import os
import warnings

if os.environ.get('CACHANNEL_BACKEND') == 'caffi':
    from caffi.ca import *
    from caffi.macros import *
    from caffi.constants import *
else:
    try:
        from ._ca import *
    except:
        warnings.warn("c extension is not available, trying caffi as fallback", RuntimeWarning)
        from caffi.ca import *
        from caffi.macros import *
        from caffi.constants import *




"""
This module defines the epicsPV class, which adds additional features to
Geoff Savage's CaChannel class.

Author:         Mark Rivers
Created:        Sept. 16, 2002.
Modifications:

- Mar. 25, 2014 Xiaoqiang Wang

  - Fix the call sequence inside getCallback

- Mar. 7, 2017 Xiaoqiang Wang

  - Reformat the docstring and code indent.

"""
from CaChannel import ca, CaChannel


class epicsPV(CaChannel):
    """
    This class subclasses CaChannel class to add the following
    features:

       - If a PV name is given then the class constructor will
         do a :meth:`CaChannel.CaChannel.searchw` by default.

       - :meth:`setMonitor` sets a generic callback routine for value change events.
         Subsequent :meth:`getw`, :meth:`getValue` or :meth:`array_get` calls will return the
         value from the most recent callback, and hence do not result in any
         network activity or latency. This can greatly improve performance.

       - :meth:`checkMonitor` returns a flag to indicate if a callback has occured
         since the last call to :meth:`checkMonitor`, :meth:`getw`, :meth:`getValue` or
         :meth:`array_get`. It can be used to increase efficiency in polling applications.

       - :meth:`getControl` reads the "control" and other information from an
         EPICS PV without having to use callbacks. In addition to the PV value,
         this will return the graphic, control and alarm limits, etc.

       - :meth:`putWait` calls :meth:`CaChannel.CaChannel.array_put_callback` and waits for the callback to
         occur before it returns. This allows programs to wait for record being processed
         synchronously and without user-written callbacks.
    """

    def __init__(self, pvName=None, wait=False):
        """
        Create an EPICS channel if pvName is specified, and optionally wait for connection.

        :param str pvName: An optional name of an EPICS Process Variable.
        :param bool wait: If wait is True and pvName is not None then this constructor will do a
                          :meth:`CaChannel.CaChannel.searchw` on the PV. If wait is False and pvName ist not None then
                          this constructor will do a :meth:`CaChannel.CaChannel.search` on the PV, and the user
                          must subsequently do a :meth:`CaChannel.CaChannel.pend_io` on this
                          or another :class:`epicsPV` or :class:`CaChannel` object.
        """
        # Invoke the base class initialization
        self.callBack = callBack()
        CaChannel.__init__(self)
        if pvName is not None:
            if wait:
                self.searchw(pvName)
            else:
                self.search(pvName)

    def setMonitor(self):
        """
        Sets a generic callback routine for value change events.

        Subsequent :meth:`getw`, :meth:`getValue` or :meth:`array_get` calls will return the
        value from the most recent callback, do not result in any network
        latency. This can greatly improve efficiency.
        """
        if self.callBack.monitorState != 0:
            return

        self.add_masked_array_event(None, None, ca.DBE_VALUE,
                                    getCallback, self.callBack)
        self.callBack.monitorState = 1

    def clearMonitor(self):
        """
        Cancels the effect of a previous call to :meth:`setMonitor`.

        Subsequent :meth:`getw`, :meth:`getValue` or :meth:`array_get` calls will no longer
        return the value from the most recent callback, but will actually result
        in channel access calls.
        """
        self.clear_event()
        self.callBack.monitorState = 0

    def checkMonitor(self):
        """
        Returns 1 to indicate if a value callback has occured
        since the last call to :meth:`checkMonitor`, :meth:`getw`, :meth:`getValue` or
        :meth:`array_get`, indicating that a new value is available.
        Returns 0 if no such callback has occurred.
        It can be used to increase efficiency in polling applications.
        """
        # This should be self.poll(), but that is generating errors
        self.pend_event(.0001)
        m = self.callBack.newMonitor
        self.callBack.newMonitor = 0
        return m

    def getControl(self, req_type=None, count=None, wait=1, poll=.01):
        """
        Provides a method to read the "control" and other information from an
        EPICS PV without having to use callbacks.

        It calls :meth:`CaChannel.CaChannel.CaChannel.array_get_callback`
        with a database request type of ca.dbf_type_to_DBR_CTRL(req_type).
        In addition to the PV value, this will return the graphic, control and
        alarm limits, etc.

        :param int req_type: request type. Default to field type.
        :param int count: number of elements. Default to native element count.
        :param bool wait: If this keyword is 1 (the default) then this routine waits for
                          the callback before returning. If this keyword is 0 then it is
                          the user's responsibility to wait or check for the callback
                          by calling :meth:`checkMonitor`.
        :param float poll: The timeout for :meth:`CaChannel.CaChannel.pend_event` calls, waiting for the callback
                           to occur. Shorter times reduce the latency at the price of CPU cycles.

        >>> pv = epicsPV('13IDC:m1')
        >>> pv.getControl()
        >>> for field in dir(pv.callBack):
        ...    print field, ':', getattr(pv.callBack, field)
            chid : _bfffec34_chid_p
            count : 1
            monitorState : 0
            newMonitor : 1
            putComplete : 0
            pv_loalarmlim : 0.0
            pv_loctrllim : -22.0
            pv_lodislim : -22.0
            pv_lowarnlim : 0.0
            pv_precision : 4
            pv_riscpad0 : 256
            pv_severity : 0
            pv_status : 0
            pv_units : mm
            pv_upalarmlim : 0.0
            pv_upctrllim : 28.0
            pv_updislim : 28.0
            pv_upwarnlim : 0.0
            pv_value : -15.0
            status : 1
            type : 34
        """
        if req_type is None: req_type = self.field_type()
        if wait: self.callBack.newMonitor = 0
        self.array_get_callback(ca.dbf_type_to_DBR_CTRL(req_type),
                                count, getCallback, self.callBack)
        if wait:
            while self.callBack.newMonitor == 0:
                self.pend_event(poll)

    def array_get(self, req_type=None, count=None, **keywords):
        """
        If :meth:`setMonitor` has not been called then this function simply calls
        :meth:`CaChannel.CaChannel.array_get`.  If :meth:`setMonitor` has been called then it calls
        :meth:`CaChannel.CaChannel.pend_event` with a very short timeout, and then returns the
        PV value from the last callback.
        """
        if self.callBack.monitorState != 0:
            # This should be self.poll(), but that is generating errors
            self.pend_event(.0001)
        if self.callBack.monitorState == 2:
            self.callBack.newMonitor = 0
            return self.callBack.pv_value
        else:
            return CaChannel.array_get(self, req_type, count, **keywords)

    def getw(self, req_type=None, count=None, **keywords):
        """
        If :meth:`setMonitor` has not been called then this function simply calls
        :meth:`CaChannel.CaChannel.getw`.  If :meth:`setMonitor` has been called then it calls
        :meth:`CaChannel.CaChannel.pend_event` with a very short timeout, and then returns the
        PV value from the last callback.
        """
        if self.callBack.monitorState != 0:
            # This should be self.poll(), but that is generating errors
            self.pend_event(.0001)
        if self.callBack.monitorState == 2:
            self.callBack.newMonitor = 0
            if count is None:
                return self.callBack.pv_value
            else:
                return self.callBack.pv_value[0:count]
        else:
            return CaChannel.getw(self, req_type, count, **keywords)

    def getValue(self):
        """
        If :meth:`setMonitor` has not been called then this function simply calls
        :meth:`CaChannel.CaChannel.getValue`.  If setMonitor has been called then it calls
        :meth:`CaChannel.CaChannel.pend_event` with a very short timeout, and then returns the
        PV value from the last callback.
        """
        if self.callBack.monitorState != 0:
            # This should be self.poll(), but that is generating errors
            self.pend_event(.0001)
        if self.callBack.monitorState == 2:
            self.callBack.newMonitor = 0
            return self.callBack.pv_value
        else:
            return CaChannel.getValue(self)

    def putWait(self, value, req_type=None, count=None, poll=.01):
        """
        Calls :meth:`CaChannel.CaChannel.array_put_callback` and waits for the callback to
        occur before it returns. This allows programs to wait for record being processed
        without having to handle asynchronous callbacks.

        :param value:      data to be written. For multiple values use a list or tuple
        :param req_type:   database request type (``ca.DBR_XXXX``). Defaults to be the native data type.
        :param int count:  number of data values to write. Defaults to be the native count.
        :param float poll: The timeout for :meth:`CaChannel.CaChannel.pend_event` calls, waiting for the callback to occur.
                           Shorter times reduce the latency at the price of CPU cycles.
        """
        self.callBack.putComplete = 0
        self.array_put_callback(value, req_type, count, putCallBack, self.callBack)
        while self.callBack.putComplete == 0:
            self.pend_event(poll)

class callBack:
    """
    This class is used by the epicsPV class to handle callbacks.  It is required
    to avoid circular references to the epicsPV object itself when dealing with
    callbacks, in order to allow the CaChannel destructor to be called.
    Users will only be interested in the fields that are copied to this class in
    the callback resulting from a call to epicsPV.getControl().
    """
    def __init__(self):
        self.newMonitor = 0
        self.putComplete = 0
        self.monitorState = 0
        self.pv_value = None
        # monitorState:
        #   0=not monitored
        #   1=monitor requested, but no callback yet
        #   2=monitor requested, callback has arrived


def putCallBack(epicsArgs, userArgs):
    """
    This is the generic callback function used by the epicsPV.putWait() method.
    It simply sets the callBack.putComplete flag to 1.
    """
    userArgs[0].putComplete = 1

def getCallback(epicsArgs, userArgs):
    """
    This is the generic callback function enabled by the epicsPV.setMonitor() method.
    It sets the callBack.monitorState flag to 2, indicating that a monitor has
    been received.  It copies all of the attributes in the epicsArgs dictionary
    to the callBack attribute of the epicsPV object.
    """
    for key in epicsArgs.keys():
        setattr(userArgs[0], key, epicsArgs[key])
    if userArgs[0].monitorState == 1: userArgs[0].monitorState = 2
    userArgs[0].newMonitor = 1



"""
This module provides functions similiar to those command line tools found in EPICS base,
e.g. :func:`caget`, :func:`caput`, :func:`camonitor`, :func:`cainfo`.

In those functions, :class:`CaChannel.CaChannel` objects are created implicitly and cached
in :data:`_channel_` dictionary.

>>> import time
>>> caput('catest', 1.23, wait=True)
>>> caget('catest')
1.23
>>> caput('cabo', 'Busy')
>>> caget('cabo')
1
>>> caget('cabo', as_string=True)
'Busy'
>>> caput('cawavec', 'this can be a long string')
>>> caget('cawavec', as_string=True)
'this '
>>> caput('cawave', range(4))
>>> caget('cawave', count=4)
[0.0, 1.0, 2.0, 3.0]
"""


import collections
import datetime
import itertools
import numbers
import sys
import threading

from .CaChannel import ca, CaChannel

#: channel object cache
_channels_ = {}


def _ints_to_string(integers):
    if isinstance(integers, collections.Sequence):
        stripped = itertools.takewhile(lambda x: x != 0, integers)
        if sys.hexversion < 0x03000000:
            value = ''.join([chr(c) for c in stripped])
        else:
            value = bytes(stripped).decode()
    elif isinstance(integers, numbers.Integral):
        if integers == 0:
            value = ''
        elif sys.hexversion < 0x03000000:
            value = chr(integers)
        else:
            value = bytes([integers]).decode()
    else:
        value = None

    return value


def _get_or_create_channel(name):
    """
    return the channel object associated with *name*. If nothing exists, create a new one.

    :param str name: pv name
    :return: channel object
    :rtype: :class:`CaChannel.CaChannel`
    """
    chan = _channels_.get(name)
    if chan is None:
        chan = CaChannel(name)
        chan.searchw()
        _channels_[name] = chan
    return chan


def caget(name, as_string=False, count=None):
    """
    Return PV's current value.

    For enum or char type PV, the string form is returned if *as_string* is True.
    If the PV is of multi-element array, *count* can be used to limit the number
    of elements.

    :param str name: pv name
    :param bool as_string: retrieve enum and char type as string
    :param int count: number of element to request
    :return: pv value
    """
    chan = _get_or_create_channel(name)

    req_type = ca.dbf_type_to_DBR(chan.field_type())
    if as_string and req_type == ca.DBR_ENUM:
        req_type = ca.DBR_STRING

    value = chan.getw(req_type, count)

    if as_string and req_type == ca.DBR_CHAR:
        value = _ints_to_string(value)

    return value


def caput(name, value, wait=False, timeout=None):
    """

    :param str name: pv name
    :param value: value to write
    :param bool wait: wait for completion
    :param float timeout: seconds to wait
    """
    chan = _get_or_create_channel(name)

    def put_callback(_, user_args):
        user_args[0].set()

    if wait:
        event = threading.Event()
        chan.array_put_callback(value, None, None, put_callback, event)
        chan.flush_io()
        event.wait(timeout)
    else:
        chan.putw(value)


def camonitor(name, as_string=False, count=None, callback=None):
    """
    set a *callback* to be invoked when pv value or alarm status change.

    :param str name: pv name
    :param bool as_string: retrieve enum and char type as string
    :param int count: number of element to request
    :param callback: callback function. If *None* is specified, the default callback is to print to the console.
                     If *callback* is not a valid *callable*, any previous callback is removed.

    >>> camonitor('cacalc')
    >>> time.sleep(2) # doctest: +ELLIPSIS
    cacalc ...
    >>> def monitor_callback(epics_args, _):
    ...     for k in sorted(epics_args):
    ...         print(k, epics_args[k])
    >>> camonitor('cacalc', callback=monitor_callback)
    >>> time.sleep(2) # doctest: +ELLIPSIS
    chid ...
    count 1
    pv_nseconds ...
    pv_seconds ...
    pv_severity AlarmSeverity.No
    pv_status AlarmCondition.No
    pv_value ...
    status ECA.NORMAL
    type DBR.TIME_DOUBLE
    chid ...
    >>> camonitor('cacalc', callback=())
    >>> time.sleep(2)

    """
    chan = _get_or_create_channel(name)

    req_type = ca.dbf_type_to_DBR_TIME(chan.field_type())
    if as_string and req_type == ca.DBR_TIME_ENUM:
        req_type = ca.DBR_TIME_STRING

    def monitor_callback(epics_args, _):
        stamp = epics_args['pv_seconds'] + ca.POSIX_TIME_AT_EPICS_EPOCH + epics_args['pv_nseconds'] * 1e-9
        # time.strftime does not support microseconds
        strfmt = datetime.datetime.fromtimestamp(stamp).strftime('%Y-%m-%d %H:%M:%S.%f')
        value = epics_args['pv_value']

        if as_string and req_type == ca.DBR_CHAR:
            value = _ints_to_string(value)

        print('%s %s %s' % (name, strfmt, value))

    if callback is not None and not callable(callback):
        chan.clear_event()
        chan.flush_io()
        return

    if callback is None:
        callback = monitor_callback

    chan.add_masked_array_event(req_type, count, None, callback)
    chan.flush_io()


def cainfo(name):
    """
    print pv information

    :param name: pv name

    >>> caput('cabo', 1)
    >>> cainfo('cabo') # doctest: +ELLIPSIS
    cabo
        State:          Connected
        Host:           ...
        Data type:      DBF_ENUM
        Element count:  1
        Access:         RW
        Status:         STATE
        Severity:       MINOR
        Enumerates:     ('Done', 'Busy')

    """
    chan = _get_or_create_channel(name)

    r = chan.read_access()
    w = chan.write_access()
    if not r and not w:
        access = 'No access'
    else:
        access = ''
        if r:
            access += 'R'
        if w:
            access += 'W'

    message = \
        """%s
    State:          %s
    Host:           %s
    Data type:      %s
    Element count:  %d
    Access:         %s""" % (
            name,
            ['Not connected', 'Connected'][chan.state() == ca.cs_conn],
            chan.host_name(),
            ca.dbf_text(chan.field_type()),
            chan.element_count(),
            access)

    if chan.state() == ca.cs_conn:
        ctrl = chan.getw(ca.dbf_type_to_DBR_CTRL(chan.field_type()))
        message += \
            """
    Status:         %s
    Severity:       %s""" % (ca.alarmStatusString(ctrl['pv_status']), ca.alarmSeverityString(ctrl['pv_severity']))

        if chan.field_type() == ca.DBF_ENUM:
            message += \
                """
    Enumerates:     %s""" % (ctrl['pv_statestrings'],)

    print(message)


if __name__ == '__main__':
    import time
    import doctest
    doctest.testmod()



"""
CaChannel class having identical API as of caPython/CaChannel class,
based on `caffi.ca API <https://caffi.readthedocs.io/en/latest/api.html>`.
"""
# python 2 -> 3 compatible layer
from __future__ import print_function, absolute_import
import collections
from functools import wraps
import itertools
import numbers
import sys
import traceback

import CaChannel as PACKAGE
from . import ca


class CaChannelException(Exception):
    def __init__(self, status):
        self.status = status

    def __int__(self):
        return self.status

    def __str__(self):
        return ca.message(self.status)


class CaChannel:
    """CaChannel: A Python class with identical API as of caPython/CaChannel.

    This class implements the methods to operate on channel access so that you can find
    their  C library counterparts ,
    http://www.aps.anl.gov/epics/base/R3-14/12-docs/CAref.html#Function.
    Therefore an understanding of C API helps much.

    To get started easily, convenient methods are created for often used operations,

    ==========    ======
    Operation     Method
    ==========    ======
    connect       :meth:`searchw`
    read          :meth:`getw`
    write         :meth:`putw`
    ==========    ======

    They have shorter names and default arguments. It is recommended to start with these methods.
    Study the other C alike methods when necessary.

    >>> chan = CaChannel('catest')
    >>> chan.searchw()
    >>> chan.putw(12.5)
    >>> chan.getw()
    12.5
    >>> chan.searchw('cabo')
    >>> chan.putw('Done')
    >>> chan.getw(ca.DBR_STRING)
    'Done'
    """

    __context = None

    ca_timeout = 3.0

    # A wrapper to automatically attach to default CA context
    def attach_ca_context(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if CaChannel.__context is None:
                ca.create_context(True)
                CaChannel.__context = ca.current_context()
            else:
                ca.attach_context(CaChannel.__context)
            return func(*args, **kwargs)

        return wrapper

    def __init__(self, pvName=None):
        self._pvname = pvName
        self._chid = None
        self._evid = None
        self._timeout = None
        self._dbrvalue = None
        self._callbacks = {}

    def __del__(self):
        try:
            self.clear_channel()
            self.flush_io()
        except:
            pass

    @staticmethod
    def version():
        return PACKAGE.__version__

    #
    # Class helper methods
    #
    def setTimeout(self, timeout):
        """Set the timeout for this channel object. It overrides the class timeout.

        :param float timeout:  timeout in seconds

        >>> chan = CaChannel()
        >>> chan.setTimeout(10.)
        >>> chan.getTimeout()
        10.0
        """
        if timeout >= 0 or timeout is None:
            self._timeout = timeout
        else:
            raise ValueError

    def getTimeout(self):
        """Retrieve the timeout set for this channel object.

        :return: timeout in seconds for this channel instance
        :rtype: float

        >>> chan = CaChannel()
        >>> chan.getTimeout() == CaChannel.ca_timeout
        True
        """
        if self._timeout is None:
            timeout = CaChannel.ca_timeout
        else:
            timeout = self._timeout

        return timeout

    def replace_access_rights_event(self, callback=None, user_args=None):
        """
        Install or replace the access rights state change callback handler for the specified channel.

        The callback handler is called in the following situations.

            - whenever CA connects the channel immediately before the channel's connection handler is called
            - whenever CA disconnects the channel immediately after the channel's disconnect callback is called
            - once immediately after installation if the channel is connected
            - whenever the access rights state of a connected channel changes

        When a channel is created no access rights handler is installed.

        :param callable callback: function called when access rights change. If None is given,
                                  remove the access rights event callback.
        :param user_args: user provided arguments that are passed to callback when it is invoked.

        >>> chan = CaChannel('catest')
        >>> chan.searchw()
        >>> def accessCB(epicsArgs, _):
        ...     print('read:', epicsArgs['read_access'], 'write:', epicsArgs['write_access'])
        >>> chan.replace_access_rights_event(accessCB)
        read: True write: True
        >>> chan.replace_access_rights_event() # clear the callback

        .. versionadded:: 3.0
        """
        if callable(callback):
            self._callbacks['accessCB'] = (callback, user_args)
            ca.replace_access_rights_event(self._chid, self._access_callback)
        else:
            self._callbacks['accessCB'] = None
            ca.replace_access_rights_event(self._chid)

    #
    # *************** Channel access method ***************
    #
    #
    # Connection methods
    #   search_and_connect
    #   search
    #   clear_channel
    #   change_connection_event

    @attach_ca_context
    def search_and_connect(self, pvName, callback, *user_args):
        """Attempt to establish a connection to a process variable.

        :param pvName:     process variable name
        :param callback:   function called when connection completes and connection status changes later on.
        :param user_args:  user provided arguments that are passed to callback when it is invoked.
        :type pvName:   bytes, str
        :type callback: callable
        :raises CaChannelException: if error happens

        The user arguments are returned to the user in a tuple in the callback function.
        The order of the arguments is preserved.

        Each Python callback function is required to have two arguments.
        The first argument is a tuple containing the results of the action.
        The second argument is a tuple containing any user arguments specified by *user_args*.
        If no arguments were specified then the tuple is empty.


        .. note:: All remote operation requests such as the above are accumulated (buffered) and not forwarded to
           the IOC until one of execution methods (:meth:`pend_io`, :meth:`poll`, :meth:`pend_event`, :meth:`flush_io`)
           is called. This allows several requests to be efficiently sent over the network in one message.

        >>> chan = CaChannel('catest')
        >>> def connCB(epicsArgs, _):
        ...     chid = epicsArgs[0]
        ...     connection_state = epicsArgs[1]
        ...     if connection_state == ca.CA_OP_CONN_UP:
        ...         print(ca.name(chid), "is connected")
        >>> chan.search_and_connect(None, connCB, chan)
        >>> status = chan.pend_event(3) # doctest: +SKIP
        catest is connected
        >>> chan.search_and_connect('cabo', connCB, chan)
        >>> status = chan.pend_event(3) # doctest: +SKIP
        cabo is connected
        >>> chan.clear_channel()
        """
        if self._chid is not None:
            self.clear_channel()

        if pvName is None:
            pvName = self._pvname
        else:
            self._pvname = pvName

        self._callbacks['connCB'] = (callback, user_args)

        status, self._chid = ca.create_channel(pvName, self._conn_callback)
        if status != ca.ECA_NORMAL:
            raise CaChannelException(status)

    @attach_ca_context
    def search(self, pvName=None):
        """Attempt to establish a connection to a process variable.

        :param pvName: process variable name
        :type pvName: bytes, str
        :raises CaChannelException: if error happens

        .. note:: All remote operation requests such as the above are accumulated (buffered) and not forwarded to
           the IOC until one of execution methods (:meth:`pend_io`, :meth:`poll`, :meth:`pend_event`, :meth:`flush_io`)
           is called. This allows several requests to be efficiently sent over the network in one message.

        >>> chan = CaChannel()
        >>> chan.search('catest')
        >>> status = chan.pend_io()
        >>> chan.state()
        <ChannelState.CONN: 2>
        """
        if self._chid is not None:
            self.clear_channel()

        if pvName is None:
            pvName = self._pvname
        else:
            self._pvname = pvName

        status, self._chid = ca.create_channel(pvName)
        if status != ca.ECA_NORMAL:
            raise CaChannelException(status)

    def clear_channel(self):
        """Close a channel created by one of the search functions.

        Clearing a channel does not cause its connection handler to be called.
        Clearing a channel does remove any monitors registered for that channel.
        If the channel is currently connected then resources are freed only some
        time after this request is flushed out to the server.

        .. note:: All remote operation requests such as the above are accumulated (buffered) and not forwarded to
           the IOC until one of execution methods (:meth:`pend_io`, :meth:`poll`, :meth:`pend_event`, :meth:`flush_io`)
           is called. This allows several requests to be efficiently sent over the network in one message.

        """
        if self._chid is not None:
            chid = self._chid
            self._chid = None
            self._dbrvalue = None
            self._callbacks = {}

            self.clear_event()
            ca.clear_channel(chid)

    def change_connection_event(self, callback, *user_args):
        """Change the connection callback function

        :param callback:   function called when connection completes and connection status changes later on.
                           The previous connection callback will be replaced. If an invalid callback is given,
                           no connection callback will be used.
        :param user_args:  user provided arguments that are passed to callback when it is invoked.
        :type callback:    callable

        >>> chan = CaChannel('catest')
        >>> chan.search() # connect without callback
        >>> def connCB(epicsArgs, _):
        ...     chid = epicsArgs[0]
        ...     connection_state = epicsArgs[1]
        ...     if connection_state == ca.CA_OP_CONN_UP:
        ...         print(ca.name(chid), "is connected")
        >>> chan.change_connection_event(connCB) # install connection callback
        >>> status = chan.pend_event(3) # doctest: +SKIP
        catest is connected
        >>> chan.change_connection_event(None) # remove connection callback

        """

        if callable(callback):
            self._callbacks['connCB'] = (callback, user_args)
            status = ca.change_connection_event(self._chid, self._conn_callback)
        else:
            self._callbacks['connCB'] = None
            status = ca.change_connection_event(self._chid)

        if status != ca.ECA_NORMAL:
            raise CaChannelException(status)

    #
    # Write methods
    #   array_put
    #   array_put_callback
    #

    def array_put(self, value, req_type=None, count=None):
        """Write a value or array of values to a channel

        :param value:    data to be written. For multiple values use a list or tuple
        :param req_type: database request type (``ca.DBR_XXXX``). Defaults to be the native data type.
        :param count:    number of data values to write. Defaults to be the native count.
        :type value: int, float, bytes, str, list, tuple, array
        :type req_type: int, None
        :type count: int, None

        .. note:: All remote operation requests such as the above are accumulated (buffered) and not forwarded to
           the IOC until one of execution methods (:meth:`pend_io`, :meth:`poll`, :meth:`pend_event`, :meth:`flush_io`)
           is called. This allows several requests to be efficiently sent over the network in one message.

        >>> chan = CaChannel('catest')
        >>> chan.searchw()
        >>> chan.array_put(123)
        >>> chan.flush_io()
        >>> chan.getw()
        123.0
        >>> chan = CaChannel('cabo')
        >>> chan.searchw()
        >>> chan.array_put('Busy', ca.DBR_STRING)
        >>> chan.flush_io()
        >>> chan.getw()
        1
        >>> chan = CaChannel('cawave')
        >>> chan.searchw()
        >>> chan.array_put([1,2,3])
        >>> chan.flush_io()
        >>> chan.getw()
        [1.0, 2.0, 3.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        >>> chan.getw(count=3, use_numpy=True)
        array([ 1.,  2.,  3.])
        >>> chan = CaChannel('cawavec')
        >>> chan.searchw()
        >>> chan.array_put('1234',count=3)
        >>> chan.flush_io()
        >>> chan.getw(count=4)
        [49, 50, 51, 0]
        """
        status = ca.put(self._chid, value, req_type, count)
        if status != ca.ECA_NORMAL:
            raise CaChannelException(status)

    def array_put_callback(self, value, req_type, count, callback, *user_args):
        """Write a value or array of values to a channel and execute the user
        supplied callback after the put has completed.

        :param value:           data to be written. For multiple values use a list or tuple.
        :param req_type:        database request type (``ca.DBR_XXXX``). Defaults to be the native data type.
        :param count:           number of data values to write, Defaults to be the native count.
        :param callback:        function called when the write is completed.
        :param user_args:       user provided arguments that are passed to callback when it is invoked.
        :type value: int, float, bytes, str, list, tuple, array
        :type req_type: int, None
        :type count: int, None
        :type callback: callable
        :raises CaChannelException: if error happens

        Each Python callback function is required to have two arguments.
        The first argument is a dictionary containing the results of the action.

        =======  =======    =======
        field    type       comment
        =======  =======    =======
        chid     capsule    channels id structure
        type     int        database request type (ca.DBR_XXXX)
        count    int        number of values to transfered
        status   int        CA status return code (ca.ECA_XXXX)
        =======  =======    =======

        The second argument is a tuple containing any user arguments specified by *user_args*.
        If no arguments were specified then the tuple is empty.

        .. note:: All remote operation requests such as the above are accumulated (buffered) and not forwarded to
           the IOC until one of execution methods (:meth:`pend_io`, :meth:`poll`, :meth:`pend_event`, :meth:`flush_io`)
           is called. This allows several requests to be efficiently sent over the network in one message.

        >>> def putCB(epicsArgs, _):
        ...     print(ca.name(epicsArgs['chid']), 'put completed')
        >>> chan = CaChannel('catest')
        >>> chan.searchw()
        >>> chan.array_put_callback(145, None, None, putCB)
        >>> status = chan.pend_event(1)
        catest put completed
        >>> chan = CaChannel('cabo')
        >>> chan.searchw()
        >>> chan.array_put_callback('Busy', ca.DBR_STRING, None, putCB)
        >>> status = chan.pend_event(1)
        cabo put completed
        >>> chan = CaChannel('cawave')
        >>> chan.searchw()
        >>> chan.array_put_callback([1,2,3], None, None, putCB)
        >>> status = chan.pend_event(1)
        cawave put completed
        >>> chan = CaChannel('cawavec')
        >>> chan.searchw()
        >>> chan.array_put_callback('123', None, None, putCB)
        >>> status = chan.pend_event(1)
        cawavec put completed
        """
        status = ca.put(self._chid, value, req_type, count, lambda epics_args: callback(epics_args, user_args))
        if status != ca.ECA_NORMAL:
            raise CaChannelException(status)
    #
    # Read methods
    #   getValue
    #   array_get
    #   array_get_callback
    #

    def getValue(self):
        """
        Return the value(s) after :meth:`array_get` has completed.

        :return: the value returned from the last array_get
        :rtype: int, float, str, list, array, dict

        If the *req_type* was not a plain type, the returned value is of dict type. It contains the same keys as in :meth:`array_get_callback`.

        .. seealso:: :meth:`array_get`

        >>> chan = CaChannel('cabo')
        >>> chan.searchw()
        >>> chan.putw(1)
        >>> chan.array_get(req_type=ca.DBR_CTRL_ENUM)
        >>> chan.pend_io()
        >>> for k,v in sorted(chan.getValue().items()):
        ...    print(k, v)
        pv_nostrings 2
        pv_severity AlarmSeverity.Minor
        pv_statestrings ('Done', 'Busy')
        pv_status AlarmCondition.State
        pv_value 1
        """
        if self._dbrvalue is None:
            return

        dbrvalue = self._dbrvalue.get()
        if isinstance(dbrvalue, dict):
            value = {}
            CaChannel._format_value(dbrvalue, value, self._dbrvalue.use_numpy)
        else:
            value = dbrvalue

        if not self._dbrvalue.use_numpy:
            if isinstance(value, dict):
                if hasattr(value['pv_value'], 'tolist'):
                    value['pv_value'] = value['pv_value'].tolist()
            else:
                if hasattr(value, 'tolist'):
                    value = value.tolist()
        return value

    def array_get(self, req_type=None, count=None, **keywords):
        """Read a value or array of values from a channel.

        The new value is not available until a subsequent :meth:`pend_io` returns :data:`ca.ECA_NORMAL`.
        Then it can be retrieved by a call to :meth:`getValue`.

        :param req_type:    database request type (:data:`ca.DBR_XXXX`). Defaults to be the native data type.
        :param count:       number of data values to read, Defaults to be the native count.
        :param keywords:    optional arguments assigned by keywords

                            ===========   =====
                            keyword       value
                            ===========   =====
                            use_numpy     True if waveform should be returned as numpy array.
                                          Default :data:`CaChannel.USE_NUMPY`.
                            ===========   =====
        :type req_type: int, None
        :type count: int, None
        :raises CaChannelException: if error happens

        .. note:: All remote operation requests such as the above are accumulated (buffered) and not forwarded to
           the IOC until one of execution methods (:meth:`pend_io`, :meth:`poll`, :meth:`pend_event`, :meth:`flush_io`)
           is called. This allows several requests to be efficiently sent over the network in one message.

        .. seealso:: :meth:`getValue`

        >>> chan = CaChannel('catest')
        >>> chan.searchw()
        >>> chan.putw(123)
        >>> chan.array_get()
        >>> chan.pend_io()
        >>> chan.getValue()
        123.0
        """
        use_numpy = keywords.get('use_numpy', PACKAGE.USE_NUMPY)
        status, self._dbrvalue = ca.get(self._chid, req_type, count, None, use_numpy)
        if status != ca.ECA_NORMAL:
            raise CaChannelException(status)

    def array_get_callback(self, req_type, count, callback, *user_args, **keywords):
        """Read a value or array of values from a channel and execute the user
        supplied callback after the get has completed.

        :param req_type:    database request type (``ca.DBR_XXXX``). Defaults to be the native data type.
        :param count:       number of data values to read, Defaults to be the native count.
        :param callback:    function called when the get is completed.
        :param user_args:   user provided arguments that are passed to callback when it is invoked.
        :param keywords:    optional arguments assigned by keywords

                            ===========   =====
                            keyword       value
                            ===========   =====
                            use_numpy     True if waveform should be returned as numpy array.
                                          Default :data:`CaChannel.USE_NUMPY`.
                            ===========   =====
        :type req_type:     int, None
        :type count:        int, None
        :type callback:     callable

        :raises CaChannelException: if error happens

        Each Python callback function is required to have two arguments.
        The first argument is a dictionary containing the results of the action.

        +-----------------+---------------+------------------------------------+-------------------------+---------------+-------------+---------------+
        | field           |  type         |  comment                           |       request type                                                    |
        |                 |               |                                    +----------+--------------+---------------+-------------+---------------+
        |                 |               |                                    | DBR_XXXX | DBR_STS_XXXX | DBR_TIME_XXXX | DBR_GR_XXXX | DBR_CTRL_XXXX |
        +=================+===============+====================================+==========+==============+===============+=============+===============+
        | chid            |   int         |   channels id number               |    X     |       X      |     X         |   X         | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | type            |   int         |   database request type            |    X     |       X      |     X         |   X         | X             |
        |                 |               |   (ca.DBR_XXXX)                    |          |              |               |             |               |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | count           |   int         |   number of values to transfered   |    X     |       X      |     X         |   X         | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | status          |   int         |   CA status return code            |    X     |       X      |     X         |   X         | X             |
        |                 |               |   (ca.ECA_XXXX)                    |          |              |               |             |               |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_value        |               |   PV value                         |    X     |       X      |     X         |   X         | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_status       |   int         |   PV alarm status                  |          |       X      |     X         |   X         | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_severity     |   int         |   PV alarm severity                |          |       X      |     X         |   X         | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_seconds      |   int         |   seconds part of timestamp        |          |              |     X         |             |               |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_nseconds     |   int         |   nanoseconds part of timestamp    |          |              |     X         |             |               |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_nostrings    |   int         |   ENUM PV's number of states       |          |              |               |   X         | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_statestrings |   string list |   ENUM PV's states string          |          |              |               |   X         | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_units        |   string      |   units                            |          |              |               |   X         | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_precision    |   int         |   precision                        |          |              |               |   X         | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_updislim     |   float       |   upper display limit              |          |              |               |   X         | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_lodislim     |   float       |   lower display limit              |          |              |               |   X         | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_upalarmlim   |   float       |   upper alarm limit                |          |              |               |   X         | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_upwarnlim    |   float       |   upper warning limit              |          |              |               |   X         | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_loalarmlim   |   float       |   lower alarm limit                |          |              |               |   X         | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_lowarnlim    |   float       |   lower warning limit              |          |              |               |   X         | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_upctrllim    |   float       |   upper control limit              |          |              |               |             | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+
        | pv_loctrllim    |   float       |   lower control limit              |          |              |               |             | X             |
        +-----------------+---------------+------------------------------------+----------+--------------+---------------+-------------+---------------+

        The second argument is a tuple containing any user arguments specified by *user_args*.
        If no arguments were specified then the tuple is empty.

        .. note:: All remote operation requests such as the above are accumulated (buffered)
           and not forwarded to the IOC until one of execution methods (:meth:`pend_io`, :meth:`poll`, :meth:`pend_event`, :meth:`flush_io`)
           is called. This allows several requests to be efficiently sent over the network in one message.

        >>> def getCB(epicsArgs, _):
        ...     for item in sorted(epicsArgs.keys()):
        ...         if item.startswith('pv_'):
        ...             print(item,epicsArgs[item])
        >>> chan = CaChannel('catest')
        >>> chan.searchw()
        >>> chan.putw(145)
        >>> chan.array_get_callback(ca.DBR_CTRL_DOUBLE, 1, getCB)
        >>> status = chan.pend_event(1)
        pv_loalarmlim -20.0
        pv_loctrllim 0.0
        pv_lodislim -20.0
        pv_lowarnlim -10.0
        pv_precision 4
        pv_severity AlarmSeverity.Major
        pv_status AlarmCondition.HiHi
        pv_units mm
        pv_upalarmlim 20.0
        pv_upctrllim 0.0
        pv_updislim 20.0
        pv_upwarnlim 10.0
        pv_value 145.0
        >>> chan = CaChannel('cabo')
        >>> chan.searchw()
        >>> chan.putw(0)
        >>> chan.array_get_callback(ca.DBR_CTRL_ENUM, 1, getCB)
        >>> status = chan.pend_event(1)
        pv_nostrings 2
        pv_severity AlarmSeverity.No
        pv_statestrings ('Done', 'Busy')
        pv_status AlarmCondition.No
        pv_value 0
        """
        use_numpy = keywords.get('use_numpy', PACKAGE.USE_NUMPY)
        self._callbacks['getCB'] = (callback, user_args, use_numpy)
        status, _ = ca.get(self._chid, req_type, count, self._get_callback, use_numpy)
        if status != ca.ECA_NORMAL:
            raise CaChannelException(status)

    #
    # Monitor methods
    #   add_masked_array_event
    #   clear_event
    #
    def add_masked_array_event(self, req_type, count, mask, callback, *user_args, **keywords):
        """Specify a callback function to be executed whenever changes occur to a PV.

        Creates a new event id and stores it on self.__evid.  Only one event registered per CaChannel object.
        If an event is already registered the event is cleared before registering a new event.

        :param req_type:    database request type (``ca.DBR_XXXX``). Defaults to be the native data type.
        :param count:       number of data values to read, Defaults to be the native count.
        :param mask:        logical or of ``ca.DBE_VALUE``, ``ca.DBE_LOG``, ``ca.DBE_ALARM``.
                            Defaults to be ``ca.DBE_VALUE|ca.DBE_ALARM``.
        :param callback:    function called when the get is completed.
        :param user_args:   user provided arguments that are passed to callback when it is invoked.
        :param keywords:    optional arguments assigned by keywords

                            ===========   =====
                            keyword       value
                            ===========   =====
                            use_numpy     True if waveform should be returned as numpy array.
                                          Default :data:`CaChannel.USE_NUMPY`.
                            ===========   =====
        :type req_type: int, None
        :type count: int, None
        :type mask: int, None
        :type callback: callable

        :raises CaChannelException: if error happens

        .. note:: All remote operation requests such as the above are accumulated (buffered) and not forwarded to
           the IOC until one of execution methods (:meth:`pend_io`, :meth:`poll`, :meth:`pend_event`, :meth:`flush_io`)
           is called. This allows several requests to be efficiently sent over the network in one message.

        >>> def eventCB(epicsArgs, _):
        ...     print('pv_value', epicsArgs['pv_value'])
        ...     print('pv_status', epicsArgs['pv_status'])
        ...     print('pv_severity', epicsArgs['pv_severity'])
        >>> chan = CaChannel('cabo')
        >>> chan.searchw()
        >>> chan.putw(1)
        >>> chan.add_masked_array_event(ca.DBR_STS_ENUM, None, None, eventCB)
        >>> status = chan.pend_event(1)
        pv_value 1
        pv_status AlarmCondition.State
        pv_severity AlarmSeverity.Minor
        >>> chan.add_masked_array_event(ca.DBR_STS_STRING, None, None, eventCB)
        >>> status = chan.pend_event(1)
        pv_value Busy
        pv_status AlarmCondition.State
        pv_severity AlarmSeverity.Minor
        >>> chan.clear_event()
        """
        if self._evid is not None:
            self.clear_event()
            self.flush_io()

        if mask is None:
            mask = ca.DBE_ALARM | ca.DBE_VALUE

        if req_type is None:
            req_type = self.field_type()

        if count is None:
            count = self.element_count()

        use_numpy = keywords.get('use_numpy', PACKAGE.USE_NUMPY)
        self._callbacks['eventCB'] = (callback, user_args, use_numpy)

        status, self._evid = ca.create_subscription(self._chid, self._event_callback, req_type, count, mask, use_numpy)

        if status != ca.ECA_NORMAL:
            raise CaChannelException(status)

    def clear_event(self):
        """Remove previously installed callback function.

        .. note:: All remote operation requests such as the above are accumulated (buffered) and not forwarded to
           the IOC until one of execution methods (:meth:`pend_io`, :meth:`poll`, :meth:`pend_event`, :meth:`flush_io`)
           is called. This allows several requests to be efficiently sent over the network in one message.
        """
        if self._evid is not None:
            status = ca.clear_subscription(self._evid)
            self._evid = None
            if status != ca.ECA_NORMAL:
                raise CaChannelException(status)

    #
    # Execute methods
    #   pend_io
    #   pend_event
    #   poll
    #   flush_io
    #

    @attach_ca_context
    def pend_io(self, timeout=None):
        """
        Flush the send buffer and wait until outstanding queries (:meth:`search`, :meth:`array_get`) complete
        or the specified timeout expires.

        :param float timeout: seconds to wait
        :raises CaChannelException: if timeout or other error happens

        """
        if timeout is None:
            timeout = self.getTimeout()
        status = ca.pend_io(float(timeout))
        if status != ca.ECA_NORMAL:
            raise CaChannelException(status)

    @attach_ca_context
    def pend_event(self, timeout=None):
        """Flush the send buffer and process background activity (connect/get/put/monitor callbacks) for ``timeout`` seconds.

        It will not return before the specified timeout expires and all unfinished channel access labor has been processed.

        :param float timeout: seconds to wait
        :return: :data:`ca.ECA_TIMEOUT`
        """
        if timeout is None:
            timeout = 0.1
        status = ca.pend_event(timeout)
        # status is always ECA_TIMEOUT
        return status

    @attach_ca_context
    def poll(self):
        """Flush the send buffer and execute any outstanding background activity.

        :return: :data:`ca.ECA_TIMEOUT`

        .. note:: It is an alias to ``pend_event(1e-12)``.
        """
        status = ca.poll()
        # status is always ECA_TIMEOUT
        return status

    @attach_ca_context
    def flush_io(self):
        """
        Flush the send buffer and does not execute outstanding background activity.

        :raises CaChannelException: if error happens
        """
        status = ca.flush_io()
        if status != ca.ECA_NORMAL:
            raise CaChannelException(status)

    #
    # Channel Access Macros
    #   field_type
    #   element_count
    #   name
    #   state
    #   host_name
    #   read_access
    #   write_access
    #

    def field_type(self):
        """
        Native type of the PV in the server, :data:`ca.DBF_XXXX`.

        >>> chan = CaChannel('catest')
        >>> chan.searchw()
        >>> ftype = chan.field_type()
        >>> ftype
        <DBF.DOUBLE: 6>
        >>> ca.dbf_text(ftype)
        'DBF_DOUBLE'
        >>> ca.DBF_DOUBLE == ftype
        True
        """
        return ca.field_type(self._chid)

    def element_count(self):
        """
        Maximum array element count of the PV in the server.

        >>> chan = CaChannel('catest')
        >>> chan.searchw()
        >>> chan.element_count()
        1
        """
        return ca.element_count(self._chid)

    def name(self):
        """
        Channel name specified when the channel was created.

        >>> chan = CaChannel('catest')
        >>> chan.searchw()
        >>> chan.name()
        'catest'
        """
        return ca.name(self._chid)

    def state(self):
        """
        Current state of the CA connection.

            ==================  =====  =============
            States              Value  Meaning
            ==================  =====  =============
            ca.cs_never_conn    0      PV not found
            ca.cs_prev_conn     1      PV was found but unavailable
            ca.cs_conn          2      PV was found and available
            ca.cs_closed        3      PV not closed
            ca.cs_never_search  4      PV not searched yet
            ==================  =====  =============

        >>> chan = CaChannel('catest')
        >>> chan.state()
        <ChannelState.NEVER_SEARCH: 4>
        >>> chan.searchw()
        >>> chan.state()
        <ChannelState.CONN: 2>
        """
        return ca.state(self._chid)

    def host_name(self):
        """
        Host name that hosts the process variable.

        >>> chan = CaChannel('catest')
        >>> chan.searchw()
        >>> host_name = chan.host_name()

        """
        return ca.host_name(self._chid)

    def read_access(self):
        """Access right to read the channel.

        :return: True if the channel can be read, False otherwise.

        >>> chan = CaChannel('catest')
        >>> chan.searchw()
        >>> chan.read_access()
        True
         """
        return ca.read_access(self._chid)

    def write_access(self):
        """Access right to write the channel.

        :return: True if the channel can be written, False otherwise.

        >>> chan = CaChannel('catest')
        >>> chan.searchw()
        >>> chan.write_access()
        True
        """
        return ca.write_access(self._chid)

    #
    # Wait functions
    #
    # These functions wait for completion of the requested action.
    def searchw(self, pvName=None):
        """
        Attempt to establish a connection to a process variable.

        :param pvName: process variable name
        :type pvName: str, None
        :raises CaChannelException: if timeout or error happens

        .. note:: This method waits for connection to be established or fail with exception.

        >>> chan = CaChannel('non-exist-channel')
        >>> chan.searchw() # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        CaChannelException: User specified timeout on IO operation expired
        """
        self.search(pvName)
        timeout = self.getTimeout()
        self.pend_io(timeout)

    def putw(self, value, req_type=None):
        """
        Write a value or array of values to a channel.

        If the request type is omitted the data is written as the Python type corresponding to the native format.
        Multi-element data is specified as a tuple or a list.
        Internally the sequence is converted to a list before inserting the values into a C array.
        Access using non-numerical types is restricted to the first element in the data field.
        Mixing character types with numerical types writes bogus results but is not prohibited at this time.
        :data:`ca.DBF_ENUM` fields can be written using :data:`ca.DBR_ENUM` and :data:`ca.DBR_STRING` types.
        :data:`ca.DBR_STRING` writes of a field of type :data:`ca.DBF_ENUM` must be accompanied by a valid string
        out of the possible enumerated values.

        :param value:    data to be written. For multiple values use a list or tuple
        :param req_type: database request type (:data:`ca.DBR_XXXX`). Defaults to be the native data type.
        :type value:     int, float, bytes, str, tuple, list, array
        :type req_type:  int, None
        :raises CaChannelException: if timeout or error happens

        .. note:: This method does flush the request to the channel access server.

        >>> chan = CaChannel('catest')
        >>> chan.searchw()
        >>> chan.putw(145)
        >>> chan.getw()
        145.0
        >>> chan = CaChannel('cabo')
        >>> chan.searchw()
        >>> chan.putw('Busy', ca.DBR_STRING)
        >>> chan.getw()
        1
        >>> chan.getw(ca.DBR_STRING)
        'Busy'
        >>> chan = CaChannel('cawave')
        >>> chan.searchw()
        >>> chan.putw([1,2,3])
        >>> chan.getw(req_type=ca.DBR_LONG,count=4)
        [1, 2, 3, 0]
        >>> chan = CaChannel('cawavec')
        >>> chan.searchw()
        >>> chan.putw('123')
        >>> chan.getw(count=4)
        [49, 50, 51, 0]
        >>> chan.getw(req_type=ca.DBR_STRING)
        '123'
        >>> chan = CaChannel('cawaves')
        >>> chan.searchw()
        >>> chan.putw(['string 1','string 2'])
        >>> chan.getw()
        ['string 1', 'string 2', '']
        """
        self.array_put(value, req_type)
        self.flush_io()

    def getw(self, req_type=None, count=None, **keywords):
        """Read the value from a channel.

        If the request type is omitted the data is returned to the user as the Python type corresponding to the native format.
        Multi-element data has all the elements returned as items in a list and must be accessed using a numerical type.
        Access using non-numerical types is restricted to the first element in the data field.

        :data:`ca.DBF_ENUM` fields can be read using :data:`ca.DBR_ENUM` and :data:`ca.DBR_STRING` types.
        :data:`ca.DBR_STRING` reads of a field of type :data:`ca.DBF_ENUM` returns the string corresponding
        to the current enumerated value.

        :data:`ca.DBF_CHAR` fields can be read using :data:`ca.DBR_CHAR` and :data:`ca.DBR_STRING` types.
        :data:`ca.DBR_CHAR` returns a scalar or a sequnece of integers. :data:`ca.DBR_STRING` assumes each integer as a
        character and assemble a string.

        :param req_type:    database request type. Defaults to be the native data type.
        :param count:       number of data values to read, Defaults to be the native count.
        :param keywords:    optional arguments assigned by keywords

                            ===========   =====
                            keyword       value
                            ===========   =====
                            use_numpy     True if waveform should be returned as numpy array.
                                          Default :data:`CaChannel.USE_NUMPY`.
                            ===========   =====
        :type req_type:     int, None
        :type count:        int, None
        :return:            If *req_type* is plain request type, only the value is returned.
                            Otherwise a dict returns with information depending on the request type,
                            same as the first argument passed to user's callback by :meth:`array_get_callback`.

        :raises CaChannelException: if timeout error happens

        >>> chan = CaChannel('catest')
        >>> chan.searchw()
        >>> chan.putw(0)
        >>> value = chan.getw(ca.DBR_TIME_DOUBLE)
        >>> for k,v in sorted(value.items()): # doctest: +ELLIPSIS
        ...    print(k, v)
        pv_nseconds ...
        pv_seconds ...
        pv_severity AlarmSeverity.No
        pv_status AlarmCondition.No
        pv_value 0.0

        .. versionchanged:: 3.0
           If *req_type* is DBR_XXX_STRING for a char type PV, a string will be returned from composing
           each element as a character.

        """
        # char waveform can be requested as string
        dbr_string_to_char = {
            ca.DBR_STRING: ca.DBR_CHAR,
            ca.DBR_STS_STRING: ca.DBR_STS_CHAR,
            ca.DBR_TIME_STRING: ca.DBR_TIME_CHAR,
            ca.DBR_GR_STRING: ca.DBR_GR_CHAR,
            ca.DBR_CTRL_STRING: ca.DBR_CTRL_CHAR
        }
        char_as_string = False
        if self.field_type() == ca.DBF_CHAR and req_type in dbr_string_to_char:
            req_type = dbr_string_to_char[req_type]
            char_as_string = True

        self.array_get(req_type, count, **keywords)
        timeout = self.getTimeout()
        self.pend_io(timeout)

        value = self.getValue()

        # convert char
        if char_as_string:
            if isinstance(value, dict):
                value['pv_value'] = CaChannel._ints_to_string(value['pv_value'])
            else:
                value = CaChannel._ints_to_string(value)

        return value

    #
    # Callback functions
    #
    # These functions hook user supplied callback functions to CA extension

    def _access_callback(self, epicsArgs):
        callback = self._callbacks.get('accessCB')
        if callback is None:
            return
        callbackFunc, userArgs = callback
        try:
            callbackFunc(epicsArgs, userArgs)
        except:
            traceback.print_exc()

    def _conn_callback(self, epicsArgs):
        callback = self._callbacks.get('connCB')
        if callback is None:
            return
        callbackFunc, userArgs = callback
        epicsArgs = (epicsArgs['chid'], epicsArgs['op'])
        try:
            callbackFunc(epicsArgs, userArgs)
        except:
            traceback.print_exc()

    def _put_callback(self, epicsArgs):
        callback = self._callbacks.get('putCB')
        if callback is None:
            return
        callbackFunc, userArgs = callback
        try:
            callbackFunc(epicsArgs, userArgs)
        except:
            traceback.print_exc()

    def _get_callback(self, epicsArgs):
        callback = self._callbacks.get('getCB')
        if callback is None:
            return
        callbackFunc, userArgs, use_numpy = callback
        epicsArgs = CaChannel._format_cb_args(epicsArgs, use_numpy)
        try:
            callbackFunc(epicsArgs, userArgs)
        except:
            traceback.print_exc()

    def _event_callback(self, epicsArgs):
        callback = self._callbacks.get('eventCB')
        if callback is None:
            return
        callbackFunc, userArgs, use_numpy = callback
        epicsArgs = CaChannel._format_cb_args(epicsArgs, use_numpy)
        try:
            callbackFunc(epicsArgs, userArgs)
        except:
            traceback.print_exc()

    @staticmethod
    def _format_cb_args(args, use_numpy):
        epicsArgs = {
            'chid': args['chid'],
            'type': args['type'],
            'count': args['count'],
            'status': args['status']
        }
        CaChannel._format_value(args['value'], epicsArgs, use_numpy)

        return epicsArgs

    @staticmethod
    def _format_value(value, epicsArgs, use_numpy):
        # dbr fields get renamed for CaChannel API
        key_map = {
            'lower_alarm_limit':   'loalarmlim',
            'lower_ctrl_limit':    'loctrllim',
            'lower_disp_limit':    'lodislim',
            'lower_warning_limit': 'lowarnlim',
            'upper_alarm_limit':   'upalarmlim',
            'upper_ctrl_limit':    'upctrllim',
            'upper_disp_limit':    'updislim',
            'upper_warning_limit': 'upwarnlim',
            'no_str': 'nostrings',
            'strs':   'statestrings'
        }
        if isinstance(value, dict):
            for key in value:
                # convert stamp dict
                if key == 'stamp':
                    epicsArgs['pv_seconds'] = value['stamp']['seconds'] - ca.POSIX_TIME_AT_EPICS_EPOCH
                    epicsArgs['pv_nseconds'] = value['stamp']['nanoseconds']
                else:
                    # dbr fields get renamed and 'pv_' prefix
                    new_key = key_map.get(key, key)
                    epicsArgs['pv_' + new_key] = value[key]
        else:
            epicsArgs['pv_value'] = value

        if not use_numpy and hasattr(epicsArgs['pv_value'], 'tolist'):
            epicsArgs['pv_value'] = epicsArgs['pv_value'].tolist()

        return epicsArgs

    @staticmethod
    def _ints_to_string(integers):
        if isinstance(integers, str):
            value = integers
        elif isinstance(integers, collections.Sequence):
            stripped = itertools.takewhile(lambda c: c != 0, integers)
            if sys.hexversion < 0x03000000:
                value = ''.join([chr(c) for c in stripped])
            else:
                value = bytes(stripped).decode()
        elif isinstance(integers, numbers.Integral):
            if integers == 0:
                value = ''
            elif sys.hexversion < 0x03000000:
                value = chr(integers)
            else:
                value = bytes([integers]).decode()
        else:
            value = None

        return value

if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
This module provides support for the EPICS motor record.

Author:         Mark Rivers

Created:        Sept. 16, 2002

Modifications:

- Mar. 7, 2017 Xiaoqiang Wang

  - Reformat the docstring and code indent.
  - Use class property to expose certain fields.
"""
import time

import epicsPV


class epicsMotor:
    """
    This module provides a class library for the EPICS motor record.
    It uses the :class:`epicsPV.epicsPV` class, which is in turn a subclass of :class:`CaChannel.CaChannel`

    Certain motor record fields are exposed as class properties.
    They can be both read and written unless otherwise noted.

    ===============  ===========================  =====
    Property         Description                  Field
    ===============  ===========================  =====
    slew_speed       Slew speed or velocity       .VELO
    base_speed       Base or starting speed       .VBAS
    acceleration     Acceleration time (sec)      .ACCL
    description      Description of motor         .DESC
    resolution       Resolution (units/step)      .MRES
    high_limit       High soft limit (user)       .HLM
    low_limit        Low soft limit (user)        .LLM
    dial_high_limit  High soft limit (dial)       .DHLM
    dial_low_limit   Low soft limit (dial)        .DLLM
    backlash         Backlash distance            .BDST
    offset           Offset from dial to user     .OFF
    done_moving      1=Done, 0=Moving, read-only  .DMOV
    ===============  ===========================  =====

    >>> m = epicsMotor('13BMD:m38')
    >>> m.move(10)              # Move to position 10 in user coordinates
    >>> m.wait()                # Wait for motor to stop moving
    >>> m.move(50, dial=True)   # Move to position 50 in dial coordinates
    >>> m.wait()                # Wait for motor to stop moving
    >>> m.move(1, step=True, relative=True) # Move 1 step relative to current position
    >>> m.wait(start=True, stop=True) # Wait for motor to start, then to stop
    >>> m.stop()                # Stop moving immediately
    >>> high = m.high_limit     # Get the high soft limit in user coordinates
    >>> m.dial_high_limit = 100 # Set the high limit to 100 in dial coodinates
    >>> speed = m.slew_speed    # Get the slew speed
    >>> m.acceleration = 0.1    # Set the acceleration to 0.1 seconds
    >>> val = m.get_position()      # Get the desired motor position in user coordinates
    >>> dval = m.get_position(dial=True)# Get the desired motor position in dial coordinates
    >>> rbv = m.get_position(readback=True) # Get the actual position in user coordinates
    >>> rrbv = m.get_position(readback=True, step=True) # Get the actual motor position in steps
    >>> m.set_position(100)   # Set the current position to 100 in user coordinates
    >>> m.set_position(10000, step=True) # Set the current position to 10000 steps
    """

    class PVProperty(object):
        def __init__(self, name, readonly=False):
            self.name = name
            self.readonly = readonly

        def __get__(self, instance, owner):
            if instance is None:
                return self
            return instance.pvs[self.name].getw()

        def __set__(self, instance, value):
            if instance is None or self.readonly:
                return
            instance.pvs[self.name].putw(value)

    slew_speed = PVProperty('velo')
    base_speed = PVProperty('vbas')
    acceleration = PVProperty('accl')
    description = PVProperty('desc')
    resolution = PVProperty('mres')
    high_limit = PVProperty('hlm')
    low_limit = PVProperty('llm')
    dial_high_limit = PVProperty('dhlm')
    dial_low_limit = PVProperty('dllm')
    backlash = PVProperty('bdst')
    offset = PVProperty('off')
    done_moving = PVProperty('dmov', readonly=True)

    def __init__(self, name):
        """
        Creates a new epicsMotor instance.

        :param str name: The name of the EPICS motor record without any trailing period or field name.

        >>> m = epicsMotor('13BMD:m38')
        """
        self.pvs = {'val':  epicsPV.epicsPV(name+'.VAL',  wait=False),
                    'dval': epicsPV.epicsPV(name+'.DVAL', wait=False),
                    'rval': epicsPV.epicsPV(name+'.RVAL', wait=False),
                    'rlv':  epicsPV.epicsPV(name+'.RLV',  wait=False),
                    'rbv':  epicsPV.epicsPV(name+'.RBV',  wait=False),
                    'drbv': epicsPV.epicsPV(name+'.DRBV', wait=False),
                    'rrbv': epicsPV.epicsPV(name+'.RRBV', wait=False),
                    'dmov': epicsPV.epicsPV(name+'.DMOV', wait=False),
                    'stop': epicsPV.epicsPV(name+'.STOP', wait=False),
                    'velo': epicsPV.epicsPV(name+'.VELO', wait=False),
                    'vbas': epicsPV.epicsPV(name+'.VBAS', wait=False),
                    'accl': epicsPV.epicsPV(name+'.ACCL', wait=False),
                    'desc': epicsPV.epicsPV(name+'.DESC', wait=False),
                    'mres': epicsPV.epicsPV(name+'.MRES', wait=False),
                    'hlm':  epicsPV.epicsPV(name+'.HLM',  wait=False),
                    'llm':  epicsPV.epicsPV(name+'.LLM',  wait=False),
                    'dhlm': epicsPV.epicsPV(name+'.DHLM', wait=False),
                    'dllm': epicsPV.epicsPV(name+'.DLLM', wait=False),
                    'bdst': epicsPV.epicsPV(name+'.BDST', wait=False),
                    'set':  epicsPV.epicsPV(name+'.SET',  wait=False),
                    'lvio': epicsPV.epicsPV(name+'.LVIO', wait=False),
                    'lls':  epicsPV.epicsPV(name+'.LLS',  wait=False),
                    'hls':  epicsPV.epicsPV(name+'.HLS',  wait=False),
                    'off':  epicsPV.epicsPV(name+'.OFF',  wait=False)
                    }
        # Wait for all PVs to connect
        self.pvs['val'].pend_io()
        self.pvs['dmov'].setMonitor()

    def move(self, value, relative=False, dial=False, step=False, ignore_limits=False):
        """
        Moves a motor to an absolute position or relative to the current position
        in user, dial or step coordinates.

        :param float value:   The absolute position or relative amount of the move
        :param bool relative: If True, move relative to current position. The default is an absolute move.
        :param bool dial: If True, _value_ is in dial coordinates. The default is user coordinates.
        :param bool step: If True, _value_ is in steps. The default is user coordinates.
        :param bool ignore_limits: If True, suppress raising exceptions if the move results in a soft or
                                   hard limit violation.
        :raises epicsMotorException: If software limit or hard limit violation detected,
                                     unless ignore_limits=True is set.

        .. note:: The "step" and "dial" keywords are mutually exclusive.
           The "relative" keyword can be used in user, dial or step coordinates.

        >>> m=epicsMotor('13BMD:m38')
        >>> m.move(10)          # Move to position 10 in user coordinates
        >>> m.move(50, dial=True)  # Move to position 50 in dial coordinates
        >>> m.move(2, step=True, relative=True) # Move 2 steps
        """
        if dial:
            # Position in dial coordinates
            if relative:
                current = self.get_position(dial=True)
                self.pvs['dval'].putw(current+value)
            else:
                self.pvs['dval'].putw(value)

        elif step:
            # Position in steps
            if relative:
                current = self.get_position(step=True)
                self.pvs['rval'].putw(current + value)
            else:
                self.pvs['rval'].putw(value)
        else:
            # Position in user coordinates
            if relative:
                self.pvs['rlv'].putw(value)
            else:
                self.pvs['val'].putw(value)

        # Check for limit violations
        if not ignore_limits:
            self.check_limits()

    def check_limits(self):
        """
        Check wether there is a soft limit, low hard limit or high hard limit violation.

        :raises epicsMotorException: If software limit or hard limit violation detected.
        """
        limit = self.pvs['lvio'].getw()
        if limit != 0:
            raise epicsMotorException('Soft limit violation')
        limit = self.pvs['lls'].getw()
        if limit != 0:
            raise epicsMotorException('Low hard limit violation')
        limit = self.pvs['hls'].getw()
        if limit != 0:
            raise epicsMotorException('High hard limit violation')

    def stop(self):
        """
        Immediately stops a motor from moving by writing 1 to the .STOP field.

        >>> m=epicsMotor('13BMD:m38')
        >>> m.move(10)          # Move to position 10 in user coordinates
        >>> m.stop()            # Stop motor
        """
        self.pvs['stop'].putw(1)

    def get_position(self, dial=False, readback=False, step=False):
        """
        Returns the target or readback motor position in user, dial or step
        coordinates.

        :param bool readback: If True, return the readback position in the
                              desired coordinate system. The default is to return the
                              target position of the motor.
        :param bool dial:     If True, return the position in dial coordinates.
                              The default is user coordinates.
        :param bool step:     If True, return the position in steps.
                              The default is user coordinates.


        .. note:: The "step" and "dial" keywords are mutually exclusive.
           The "readback" keyword can be used in user, dial or step coordinates.

        >>> m = epicsMotor('13BMD:m38')
        >>> m.move(10)                   # Move to position 10 in user coordinates
        >>> pos_dial = m.get_position(dial=True)   # Read the target position in dial coordinates
        >>> pos_step = m.get_position(readback=True, step=True) # Read the actual position in steps
        """
        if dial:
            if readback:
                return self.pvs['drbv'].getw()
            else:
                return self.pvs['dval'].getw()
        elif step:
            if readback:
                return self.pvs['rrbv'].getw()
            else:
                return self.pvs['rval'].getw()
        else:
            if readback:
                return self.pvs['rbv'].getw()
            else:
                return self.pvs['val'].getw()

    def set_position(self, position, dial=False, step=False):
        """
        Sets the motor position in user, dial or step coordinates.

        :param float position: The new motor position
        :param bool dial:      If True, set the position in dial coordinates.
                               The default is user coordinates.
        :param bool step:      If True, set the position in steps.
                               The default is user coordinates.

        .. note:: The "step" and "dial" keywords are mutually exclusive.

        >>> m = epicsMotor('13BMD:m38')
        >>> m.set_position(10, dial=True)   # Set the motor position to 10 in dial coordinates
        >>> m.set_position(1000, step=True) # Set the motor position to 1000 steps
        """
        # Put the motor in "SET" mode
        self.pvs['set'].putw(1)
        if dial:
            self.pvs['dval'].putw(position)
        elif step:
            self.pvs['rval'].putw(position)
        else:
            self.pvs['val'].putw(position)
        # Put the motor back in "Use" mode
        self.pvs['set'].putw(0)

    def wait(self, start=False, stop=False, poll=0.01, ignore_limits=False):
        """
        Waits for the motor to start moving and/or stop moving.

        :param bool start: If True, wait for the motor to start moving.
        :param bool stop:  If True, wait for the motor to stop moving.
        :param float poll: The time to wait between reading the
                           .DMOV field of the record to see if the motor is moving.
                           The default is 0.01 seconds.
        :param bool ignore_limits: If True, suppress raising an exception if a soft or
                                   hard limit is detected.
        :raises epicsMotorException: If software limit or hard limit violation detected,
                                     unless *ignore_limits=True* is set.


        .. note:: If neither the "start" nor "stop" keywords are set then "stop"
           is set to 1, so the routine waits for the motor to stop moving.
           If only "start" is set to 1 then the routine only waits for the
           motor to start moving.
           If both "start" and "stop" are set to 1 then the routine first
           waits for the motor to start moving, and then to stop moving.

        >>> m = epicsMotor('13BMD:m38')
        >>> m.move(50)                # Move to position 50
        >>> m.wait(start=True, stop=True)   # Wait for the motor to start moving and then to stop moving
        """
        if not start and not stop:
            stop = True
        if start:
            while self.pvs['dmov'].getw() == 1:
                time.sleep(poll)
        if stop:
            while self.pvs['dmov'].getw() == 0:
                time.sleep(poll)
        if not ignore_limits:
            self.check_limits()


class epicsMotorException(Exception):
    def __init__(self, message=''):
        self.message = message

    def __str__(self):
        return self.message

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
"""
Setup file for Ca-Python using distutils package.
Python2.6 or later should be used.
"""

import os
import sys
import platform
import imp
import shutil
import subprocess
import warnings

# Use setuptools to include build_sphinx, upload/sphinx commands
try:
    from setuptools import setup, Extension
except:
    from distutils.core import setup, Extension

build_ca_ext=True
# define EPICS base path and host arch
EPICSBASE = os.environ.get("EPICS_BASE")
HOSTARCH = os.environ.get("EPICS_HOST_ARCH")
# guess from EPICS root environment variable
if not EPICSBASE:
    EPICSROOT = os.environ.get("EPICS")
    if EPICSROOT:
        EPICSBASE = os.path.join(EPICSROOT, 'base')

if not EPICSBASE or not os.path.exists(EPICSBASE) or not HOSTARCH:
    build_ca_ext = False
    warnings.warn("""
    !!! No valid EPICS_BASE, EPICS_HOST_ARCH environment variables are defined !!!
    !!!                 Switch to caffi as backend                             !!!
    """)


def create_exension():
    global EPICSBASE, HOSTARCH
    umacros = []
    macros = []
    cflags = []
    lflags = []
    dlls = []
    extra_objects = []
    libraries = ["ca", "Com"]
    CMPL = 'gcc'
    UNAME = platform.system()
    ARCH = platform.architecture()[0]
    # platform dependent libraries and macros
    if UNAME.lower() == "windows":
        UNAME = "WIN32"
        static = False
        if HOSTARCH in ['win32-x86', 'windows-x64', 'win32-x86-debug', 'windows-x64-debug']:
            dlls = ['Com.dll', 'ca.dll']
            for dll in dlls:
                dllpath = os.path.join(EPICSBASE, 'bin', HOSTARCH, dll)
                if not os.path.exists(dllpath):
                    static = True
                    break
                shutil.copy(dllpath,
                            os.path.join(os.path.dirname(os.path.abspath(__file__)), 'src', 'CaChannel'))
            macros += [('_CRT_SECURE_NO_WARNINGS', 'None'), ('EPICS_CALL_DLL', '')]
            cflags += ['/Z7']
            CMPL = 'msvc'
        if HOSTARCH in ['win32-x86-static', 'windows-x64-static'] or static:
            libraries += ['ws2_32', 'user32', 'advapi32']
            macros += [('_CRT_SECURE_NO_WARNINGS', 'None'), ('EPICS_DLL_NO', '')]
            umacros += ['_DLL']
            cflags += ['/EHsc', '/Z7']
            lflags += ['/LTCG']
            if HOSTARCH[-5:] == 'debug':
                libraries += ['msvcrtd']
                lflags += ['/NODEFAULTLIB:libcmtd.lib']
            else:
                libraries += ['msvcrt']
                lflags += ['/NODEFAULTLIB:libcmt.lib']
            CMPL = 'msvc'
        # GCC compiler
        if HOSTARCH in ['win32-x86-mingw', 'windows-x64-mingw']:
            macros += [('_MINGW', ''), ('EPICS_DLL_NO', '')]
            lflags += ['-static']
            CMPL = 'gcc'
        if HOSTARCH == 'windows-x64-mingw':
            macros += [('MS_WIN64', '')]
            CMPL = 'gcc'
    elif UNAME.lower() == "darwin":
        CMPL = 'clang'
        HOSTARCH = 'darwin-x86'
        extra_objects = [os.path.join(EPICSBASE, 'lib', HOSTARCH, 'lib%s.a' % lib) for lib in libraries]
        libraries = []
    elif UNAME.lower() == "linux":
        CMPL = 'gcc'
        extra_objects = [os.path.join(EPICSBASE, 'lib', HOSTARCH, 'lib%s.a' % lib) for lib in libraries]
        libraries = ['rt']
        if subprocess.call('nm %s | grep -q rl_' % os.path.join(EPICSBASE, 'lib', HOSTARCH, 'libCom.a'), shell=True) == 0:
            libraries += ['readline']
    else:
        print("Platform", UNAME, ARCH, " Not Supported")
        sys.exit(1)

    include_dirs = [os.path.join(EPICSBASE, "include"),
                    os.path.join(EPICSBASE, "include", "os", UNAME),
                    os.path.join(EPICSBASE, "include", "compiler", CMPL),
                    ]

    # guess numpy path
    try:
        import numpy
    except:
        warnings.warn('numpy is not present. Large array read could suffer from low efficiency.')
    else:
        macros += [('WITH_NUMPY', None)]
        include_dirs += [numpy.get_include()]

    ca_module = Extension('CaChannel._ca',
                          sources=['src/CaChannel/_ca.cpp'],
                          extra_compile_args=cflags,
                          include_dirs=include_dirs,
                          define_macros=macros,
                          undef_macros=umacros,
                          extra_link_args=lflags,
                          extra_objects=extra_objects,
                          libraries=libraries,
                          library_dirs=[os.path.join(EPICSBASE, "lib", HOSTARCH)])

    if UNAME not in ["WIN32", "Darwin", "Linux"]:
        ca_module.runtime_library_dirs = [os.path.join(EPICSBASE, "lib", HOSTARCH)]

    return [ca_module], dlls

_version = imp.load_source('_version', 'src/CaChannel/_version.py')

if build_ca_ext:
    ext_module, package_data = create_exension()
    requirements = []
    if sys.hexversion < 0x03040000:
        requirements += ['enum34']
else:
    ext_module = []
    package_data = []
    requirements = ['caffi']

setup(name="CaChannel",
      version=_version.__version__,
      author="Xiaoqiang Wang",
      author_email="xiaoqiang.wang AT psi DOT ch",
      description="CaChannel Interface to EPICS",
      long_description=open('README.rst').read(),
      url="http://pypi.python.org/pypi/cachannel",
      license="BSD",
      platforms=["Windows", "Linux", "Mac OS X"],
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Programming Language :: C',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'License :: OSI Approved :: BSD License',
          'Topic :: Scientific/Engineering',
      ],
      packages=["CaChannel"],
      package_dir={"": "src", "CaChannel": "src/CaChannel"},
      py_modules=["ca", "epicsPV", "epicsMotor"],
      ext_modules=ext_module,
      package_data={'CaChannel': package_data},
      install_requires=requirements
      )
